@@ -1,31 +1,37 @@
 /****************************************************************************
 **
-** Copyright (C) 2014 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtGui module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL21$
+** $QT_BEGIN_LICENSE:LGPL$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia. For licensing terms and
-** conditions see http://qt.digia.com/licensing. For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 or version 3 as published by the Free
-** Software Foundation and appearing in the file LICENSE.LGPLv21 and
-** LICENSE.LGPLv3 included in the packaging of this file. Please review the
-** following information to ensure the GNU Lesser General Public License
-** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
-** In addition, as a special exception, Digia gives you certain additional
-** rights. These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -125,14 +131,12 @@ static bool isValidBlockSeparator(QChar ch)
         || ch == QTextEndOfFrame;
 }
 
-#if !defined(QT_NO_DEBUG) || defined(QT_FORCE_ASSERTS)
-static bool noBlockInString(const QString &str)
+static bool noBlockInString(const QStringRef &str)
 {
     return !str.contains(QChar::ParagraphSeparator)
         && !str.contains(QTextBeginningOfFrame)
         && !str.contains(QTextEndOfFrame);
 }
-#endif
 
 bool QTextUndoCommand::tryMerge(const QTextUndoCommand &other)
 {
@@ -231,7 +235,7 @@ void QTextDocumentPrivate::clear()
 {
     Q_Q(QTextDocument);
 
-    foreach (QTextCursorPrivate *curs, cursors) {
+    for (QTextCursorPrivate *curs : qAsConst(cursors)) {
         curs->setPosition(0);
         curs->currentCharFormat = -1;
         curs->anchor = 0;
@@ -261,7 +265,7 @@ void QTextDocumentPrivate::clear()
         unreachableCharacterCount = 0;
         modifiedState = 0;
         modified = false;
-        formats = QTextFormatCollection();
+        formats.clear();
         int len = fragments.length();
         fragments.clear();
         blocks.clear();
@@ -283,7 +287,7 @@ void QTextDocumentPrivate::clear()
 
 QTextDocumentPrivate::~QTextDocumentPrivate()
 {
-    foreach (QTextCursorPrivate *curs, cursors)
+    for (QTextCursorPrivate *curs : qAsConst(cursors))
         curs->priv = 0;
     cursors.clear();
     undoState = 0;
@@ -316,7 +320,7 @@ void QTextDocumentPrivate::setLayout(QAbstractTextDocumentLayout *layout)
 void QTextDocumentPrivate::insert_string(int pos, uint strPos, uint length, int format, QTextUndoCommand::Operation op)
 {
     // ##### optimize when only appending to the fragment!
-    Q_ASSERT(noBlockInString(text.mid(strPos, length)));
+    Q_ASSERT(noBlockInString(text.midRef(strPos, length)));
 
     split(pos);
     uint x = fragments.insert_single(pos, length);
@@ -472,7 +476,7 @@ void QTextDocumentPrivate::insert(int pos, const QString &str, int format)
     if (str.size() == 0)
         return;
 
-    Q_ASSERT(noBlockInString(str));
+    Q_ASSERT(noBlockInString(QStringRef(&str)));
 
     int strPos = text.length();
     text.append(str);
@@ -490,7 +494,7 @@ int QTextDocumentPrivate::remove_string(int pos, uint length, QTextUndoCommand::
 
     Q_ASSERT(blocks.size(b) > length);
     Q_ASSERT(x && fragments.position(x) == (uint)pos && fragments.size(x) == length);
-    Q_ASSERT(noBlockInString(text.mid(fragments.fragment(x)->stringPosition, length)));
+    Q_ASSERT(noBlockInString(text.midRef(fragments.fragment(x)->stringPosition, length)));
 
     blocks.setSize(b, blocks.size(b)-length);
 
@@ -625,7 +629,7 @@ void QTextDocumentPrivate::move(int pos, int to, int length, QTextUndoCommand::O
 
         if (key+1 != blocks.position(b)) {
 //          qDebug("remove_string from %d length %d", key, X->size_array[0]);
-            Q_ASSERT(noBlockInString(text.mid(X->stringPosition, X->size_array[0])));
+            Q_ASSERT(noBlockInString(text.midRef(X->stringPosition, X->size_array[0])));
             w = remove_string(key, X->size_array[0], op);
 
             if (needsInsert) {
@@ -670,7 +674,7 @@ void QTextDocumentPrivate::remove(int pos, int length, QTextUndoCommand::Operati
     blockCursorAdjustment = true;
     move(pos, -1, length, op);
     blockCursorAdjustment = false;
-    foreach (QTextCursorPrivate *curs, cursors) {
+    for (QTextCursorPrivate *curs : qAsConst(cursors)) {
         if (curs->adjustPosition(pos, -length, op) == QTextCursorPrivate::CursorMoved) {
             curs->changed = true;
         }
@@ -1002,9 +1006,9 @@ int QTextDocumentPrivate::undoRedo(bool undo)
         bool inBlock = (
                 undoState > 0
                 && undoState < undoStack.size()
-                && undoStack[undoState].block_part
-                && undoStack[undoState-1].block_part
-                && !undoStack[undoState-1].block_end
+                && undoStack.at(undoState).block_part
+                && undoStack.at(undoState - 1).block_part
+                && !undoStack.at(undoState - 1).block_end
                 );
         if (!inBlock)
             break;
@@ -1070,12 +1074,14 @@ void QTextDocumentPrivate::appendUndoItem(const QTextUndoCommand &c)
 
 
     if (!undoStack.isEmpty() && modified) {
-        QTextUndoCommand &last = undoStack[undoState - 1];
+        const int lastIdx = undoState - 1;
+        const QTextUndoCommand &last = undoStack.at(lastIdx);
 
         if ( (last.block_part && c.block_part && !last.block_end) // part of the same block => can merge
-            || (!c.block_part && !last.block_part)) {  // two single undo items => can merge
-
-            if (last.tryMerge(c))
+            || (!c.block_part && !last.block_part) // two single undo items => can merge
+            || (c.command == QTextUndoCommand::Inserted && last.command == c.command && (last.block_part && !c.block_part))) {
+            // two sequential inserts that are not part of the same block => can merge
+            if (undoStack[lastIdx].tryMerge(c))
                 return;
         }
     }
@@ -1097,7 +1103,7 @@ void QTextDocumentPrivate::clearUndoRedoStacks(QTextDocument::Stacks stacksToCle
     bool redoCommandsAvailable = undoState != undoStack.size();
     if (stacksToClear == QTextDocument::UndoStack && undoCommandsAvailable) {
         for (int i = 0; i < undoState; ++i) {
-            QTextUndoCommand c = undoStack[undoState];
+            QTextUndoCommand c = undoStack.at(undoState);
             if (c.command & QTextUndoCommand::Custom)
                 delete c.custom;
         }
@@ -1109,7 +1115,7 @@ void QTextDocumentPrivate::clearUndoRedoStacks(QTextDocument::Stacks stacksToCle
     } else if (stacksToClear == QTextDocument::RedoStack
                && redoCommandsAvailable) {
         for (int i = undoState; i < undoStack.size(); ++i) {
-            QTextUndoCommand c = undoStack[i];
+            QTextUndoCommand c = undoStack.at(i);
             if (c.command & QTextUndoCommand::Custom)
                 delete c.custom;
         }
@@ -1119,12 +1125,12 @@ void QTextDocumentPrivate::clearUndoRedoStacks(QTextDocument::Stacks stacksToCle
     } else if (stacksToClear == QTextDocument::UndoAndRedoStacks
                && !undoStack.isEmpty()) {
         for (int i = 0; i < undoStack.size(); ++i) {
-            QTextUndoCommand c = undoStack[i];
+            QTextUndoCommand c = undoStack.at(i);
             if (c.command & QTextUndoCommand::Custom)
                 delete c.custom;
         }
         undoState = 0;
-        undoStack.resize(0);
+        undoStack.clear();
         if (emitSignals && undoCommandsAvailable)
             emitUndoAvailable(false);
         if (emitSignals && redoCommandsAvailable)
@@ -1182,8 +1188,8 @@ void QTextDocumentPrivate::endEditBlock()
         return;
 
     if (undoEnabled && undoState > 0) {
-        const bool wasBlocking = !undoStack[undoState - 1].block_end;
-        if (undoStack[undoState - 1].block_part) {
+        const bool wasBlocking = !undoStack.at(undoState - 1).block_end;
+        if (undoStack.at(undoState - 1).block_part) {
             undoStack[undoState - 1].block_end = true;
             if (wasBlocking)
                 emit document()->undoCommandAdded();
@@ -1228,13 +1234,13 @@ void QTextDocumentPrivate::finishEdit()
     }
 
     QList<QTextCursor> changedCursors;
-    foreach (QTextCursorPrivate *curs, cursors) {
+    for (QTextCursorPrivate *curs : qAsConst(cursors)) {
         if (curs->changed) {
             curs->changed = false;
             changedCursors.append(QTextCursor(curs));
         }
     }
-    foreach (const QTextCursor &cursor, changedCursors)
+    for (const QTextCursor &cursor : qAsConst(changedCursors))
         emit q->cursorPositionChanged(cursor);
 
     contentsChanged();
@@ -1280,7 +1286,7 @@ void QTextDocumentPrivate::adjustDocumentChangesAndCursors(int from, int addedOr
     if (blockCursorAdjustment)  {
         ; // postpone, will be called again from QTextDocumentPrivate::remove()
     } else {
-        foreach (QTextCursorPrivate *curs, cursors) {
+        for (QTextCursorPrivate *curs : qAsConst(cursors)) {
             if (curs->adjustPosition(from, addedOrRemoved, op) == QTextCursorPrivate::CursorMoved) {
                 curs->changed = true;
             }
@@ -1727,7 +1733,7 @@ bool QTextDocumentPrivate::ensureMaximumBlockCount()
 void QTextDocumentPrivate::aboutToRemoveCell(int from, int to)
 {
     Q_ASSERT(from <= to);
-    foreach (QTextCursorPrivate *curs, cursors)
+    for (QTextCursorPrivate *curs : qAsConst(cursors))
         curs->aboutToRemoveCell(from, to);
 }
 
