@@ -1,7 +1,7 @@
 /****************************************************************************
 **
-** Copyright (C) 2014 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the plugins of the Qt Toolkit.
 **
@@ -10,30 +10,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -45,22 +43,45 @@
 #include "qiosintegration.h"
 #include "qiosviewcontroller.h"
 #include "qiostextresponder.h"
+#include "qiosscreen.h"
 #include "qioswindow.h"
+#ifndef Q_OS_TVOS
 #include "qiosmenu.h"
+#endif
 
+#include <QtCore/qoperatingsystemversion.h>
 #include <QtGui/private/qguiapplication_p.h>
 #include <QtGui/private/qwindow_p.h>
+#include <qpa/qwindowsysteminterface_p.h>
 
 @implementation QUIView
 
++ (void)load
+{
+    if (QOperatingSystemVersion::current() < QOperatingSystemVersion(QOperatingSystemVersion::IOS, 11)) {
+        // iOS 11 handles this though [UIView safeAreaInsetsDidChange], but there's no signal for
+        // the corresponding top and bottom layout guides that we use on earlier versions. Note
+        // that we use the _will_ change version of the notification, because we want to react
+        // to the change as early was possible. But since the top and bottom layout guides have
+        // not been updated at this point we use asynchronous delivery of the event, so that the
+        // event is processed by QtGui just after iOS has updated the layout margins.
+        [[NSNotificationCenter defaultCenter] addObserverForName:UIApplicationWillChangeStatusBarFrameNotification
+            object:nil queue:[NSOperationQueue mainQueue] usingBlock:^(NSNotification *) {
+                for (QWindow *window : QGuiApplication::allWindows())
+                    QWindowSystemInterface::handleSafeAreaMarginsChanged<QWindowSystemInterface::AsynchronousDelivery>(window);
+            }
+        ];
+    }
+}
+
 + (Class)layerClass
 {
     return [CAEAGLLayer class];
 }
 
--(id)initWithQIOSWindow:(QIOSWindow *)window
+- (id)initWithQIOSWindow:(QT_PREPEND_NAMESPACE(QIOSWindow) *)window
 {
-    if (self = [self initWithFrame:toCGRect(window->geometry())])
+    if (self = [self initWithFrame:window->geometry().toCGRect()])
         m_qioswindow = window;
 
     m_accessibleElements = [[NSMutableArray alloc] init];
@@ -80,7 +101,39 @@
         if (isQtApplication())
             self.hidden = YES;
 
+#ifndef Q_OS_TVOS
         self.multipleTouchEnabled = YES;
+#endif
+
+        if (QIOSIntegration::instance()->debugWindowManagement()) {
+            static CGFloat hue = 0.0;
+            CGFloat lastHue = hue;
+            for (CGFloat diff = 0; diff < 0.1 || diff > 0.9; diff = fabs(hue - lastHue))
+                hue = drand48();
+
+            #define colorWithBrightness(br) \
+                [UIColor colorWithHue:hue saturation:0.5 brightness:br alpha:1.0].CGColor
+
+            self.layer.backgroundColor = colorWithBrightness(0.5);
+            self.layer.borderColor = colorWithBrightness(1.0);
+            self.layer.borderWidth = 1.0;
+        }
+
+#if QT_DARWIN_PLATFORM_SDK_EQUAL_OR_ABOVE(__MAC_NA, 110000, 110000, __WATCHOS_NA)
+        if (qEnvironmentVariableIsSet("QT_IOS_DEBUG_WINDOW_SAFE_AREAS")) {
+            if (__builtin_available(iOS 11, tvOS 11, *)) {
+                UIView *safeAreaOverlay = [[UIView alloc] initWithFrame:CGRectZero];
+                [safeAreaOverlay setBackgroundColor:[UIColor colorWithRed:0.3 green:0.7 blue:0.9 alpha:0.3]];
+                [self addSubview:safeAreaOverlay];
+
+                safeAreaOverlay.translatesAutoresizingMaskIntoConstraints = NO;
+                [safeAreaOverlay.topAnchor constraintEqualToAnchor:self.safeAreaLayoutGuide.topAnchor].active = YES;
+                [safeAreaOverlay.leftAnchor constraintEqualToAnchor:self.safeAreaLayoutGuide.leftAnchor].active = YES;
+                [safeAreaOverlay.rightAnchor constraintEqualToAnchor:self.safeAreaLayoutGuide.rightAnchor].active = YES;
+                [safeAreaOverlay.bottomAnchor constraintEqualToAnchor:self.safeAreaLayoutGuide.bottomAnchor].active = YES;
+            }
+        }
+#endif
     }
 
     return self;
@@ -137,7 +190,7 @@
     // from what we end up with after applying window constraints.
     QRect requestedGeometry = m_qioswindow->geometry();
 
-    QRect actualGeometry = fromCGRect(self.frame).toRect();
+    QRect actualGeometry = QRectF::fromCGRect(self.frame).toRect();
 
     // Persist the actual/new geometry so that QWindow::geometry() can
     // be queried on the resize event.
@@ -147,8 +200,8 @@
             requestedGeometry : qt_window_private(m_qioswindow->window())->geometry;
 
     QWindow *window = m_qioswindow->window();
+    qCDebug(lcQpaWindow) << m_qioswindow->window() << "new geometry is" << actualGeometry;
     QWindowSystemInterface::handleGeometryChange(window, actualGeometry, previousGeometry);
-    QWindowSystemInterface::flushWindowSystemEvents(window->inherits("QWidgetWindow") ? QEventLoop::ExcludeUserInputEvents : QEventLoop::AllEvents);
 
     if (actualGeometry.size() != previousGeometry.size()) {
         // Trigger expose event on resize
@@ -172,7 +225,7 @@
     QRegion region;
 
     if (m_qioswindow->isExposed()) {
-        QSize bounds = fromCGRect(self.layer.bounds).toRect().size();
+        QSize bounds = QRectF::fromCGRect(self.layer.bounds).toRect().size();
 
         Q_ASSERT(m_qioswindow->geometry().size() == bounds);
         Q_ASSERT(self.hidden == !m_qioswindow->window()->isVisible());
@@ -180,39 +233,46 @@
         region = QRect(QPoint(), bounds);
     }
 
+    qCDebug(lcQpaWindow) << m_qioswindow->window() << region << "isExposed" << m_qioswindow->isExposed();
     QWindowSystemInterface::handleExposeEvent(m_qioswindow->window(), region);
-    QWindowSystemInterface::flushWindowSystemEvents();
+}
+
+- (void)safeAreaInsetsDidChange
+{
+    QWindowSystemInterface::handleSafeAreaMarginsChanged(m_qioswindow->window());
 }
 
 // -------------------------------------------------------------------------
 
 - (BOOL)canBecomeFirstResponder
 {
-    return YES;
+    return !(m_qioswindow->window()->flags() & Qt::WindowDoesNotAcceptFocus);
 }
 
 - (BOOL)becomeFirstResponder
 {
-    FirstResponderCandidate firstResponderCandidate(self);
+    {
+        // Scope for the duration of becoming first responder only, as the window
+        // activation event may trigger new responders, which we don't want to be
+        // blocked by this guard.
+        FirstResponderCandidate firstResponderCandidate(self);
 
-    qImDebug() << "win:" << m_qioswindow->window() << "self:" << self
-        << "first:" << [UIResponder currentFirstResponder];
+        qImDebug() << "win:" << m_qioswindow->window() << "self:" << self
+            << "first:" << [UIResponder currentFirstResponder];
 
-    if (![super becomeFirstResponder]) {
-        qImDebug() << m_qioswindow->window()
-            << "was not allowed to become first responder";
-        return NO;
+        if (![super becomeFirstResponder]) {
+            qImDebug() << m_qioswindow->window()
+                << "was not allowed to become first responder";
+            return NO;
+        }
+
+        qImDebug() << m_qioswindow->window() << "became first responder";
     }
 
-    qImDebug() << m_qioswindow->window() << "became first responder";
-
-    if (qGuiApp->focusWindow() != m_qioswindow->window()) {
+    if (qGuiApp->focusWindow() != m_qioswindow->window())
         QWindowSystemInterface::handleWindowActivated(m_qioswindow->window());
-        QWindowSystemInterface::flushWindowSystemEvents();
-    } else {
-        qImDebug() << m_qioswindow->window()
-            << "already active, not sending window activation";
-    }
+    else
+        qImDebug() << m_qioswindow->window() << "already active, not sending window activation";
 
     return YES;
 }
@@ -247,19 +307,59 @@
     qImDebug() << m_qioswindow->window() << "resigned first responder";
 
     UIResponder *newResponder = FirstResponderCandidate::currentCandidate();
-    if ([self responderShouldTriggerWindowDeactivation:newResponder]) {
+    if ([self responderShouldTriggerWindowDeactivation:newResponder])
         QWindowSystemInterface::handleWindowActivated(0);
-        QWindowSystemInterface::flushWindowSystemEvents();
-    }
 
     return YES;
 }
 
+- (BOOL)isActiveWindow
+{
+    // Normally this is determined exclusivly by being firstResponder, but
+    // since we employ a separate first responder for text input we need to
+    // handle both cases as this view being the active Qt window.
+
+    if ([self isFirstResponder])
+        return YES;
+
+    UIResponder *firstResponder = [UIResponder currentFirstResponder];
+    if ([firstResponder isKindOfClass:[QIOSTextInputResponder class]]
+        && [firstResponder nextResponder] == self)
+        return YES;
+
+    return NO;
+}
+
 // -------------------------------------------------------------------------
 
+- (void)traitCollectionDidChange:(UITraitCollection *)previousTraitCollection
+{
+    [super traitCollectionDidChange: previousTraitCollection];
+
+    QTouchDevice *touchDevice = QIOSIntegration::instance()->touchDevice();
+    QTouchDevice::Capabilities touchCapabilities = touchDevice->capabilities();
+
+    if (QOperatingSystemVersion::current() >= QOperatingSystemVersion(QOperatingSystemVersion::IOS, 9)) {
+        if (self.traitCollection.forceTouchCapability == UIForceTouchCapabilityAvailable)
+            touchCapabilities |= QTouchDevice::Pressure;
+        else
+            touchCapabilities &= ~QTouchDevice::Pressure;
+    }
+
+    touchDevice->setCapabilities(touchCapabilities);
+}
+
+-(BOOL)pointInside:(CGPoint)point withEvent:(UIEvent *)event
+{
+    if (m_qioswindow->window()->flags() & Qt::WindowTransparentForInput)
+        return NO;
+    return [super pointInside:point withEvent:event];
+}
 
 - (void)updateTouchList:(NSSet *)touches withState:(Qt::TouchPointState)state
 {
+    bool supportsPressure = QIOSIntegration::instance()->touchDevice()->capabilities() & QTouchDevice::Pressure;
+
     foreach (UITouch *uiTouch, m_activeTouches.keys()) {
         QWindowSystemInterface::TouchPoint &touchPoint = m_activeTouches[uiTouch];
         if (![touches containsObject:uiTouch]) {
@@ -270,7 +370,8 @@
             // Touch positions are expected to be in QScreen global coordinates, and
             // as we already have the QWindow positioned at the right place, we can
             // just map from the local view position to global coordinates.
-            QPoint localViewPosition = fromCGPoint([uiTouch locationInView:self]).toPoint();
+            // tvOS: all touches start at the center of the screen and move from there.
+            QPoint localViewPosition = QPointF::fromCGPoint([uiTouch locationInView:self]).toPoint();
             QPoint globalScreenPosition = m_qioswindow->mapToGlobal(localViewPosition);
 
             touchPoint.area = QRectF(globalScreenPosition, QSize(0, 0));
@@ -280,19 +381,39 @@
             touchPoint.normalPosition = QPointF(globalScreenPosition.x() / screenSize.width(),
                                                 globalScreenPosition.y() / screenSize.height());
 
-            // We don't claim that our touch device supports QTouchDevice::Pressure,
-            // but fill in a meaningfull value in case clients use it anyways.
-            touchPoint.pressure = (state == Qt::TouchPointReleased) ? 0.0 : 1.0;
+            if (supportsPressure) {
+                // Note: iOS  will deliver touchesBegan with a touch force of 0, which
+                // we will reflect/propagate as a 0 pressure, but there is no clear
+                // alternative, as we don't want to wait for a touchedMoved before
+                // sending a touch press event to Qt, just to have a valid pressure.
+                touchPoint.pressure = uiTouch.force / uiTouch.maximumPossibleForce;
+            } else {
+                // We don't claim that our touch device supports QTouchDevice::Pressure,
+                // but fill in a meaningfull value in case clients use it anyways.
+                touchPoint.pressure = (state == Qt::TouchPointReleased) ? 0.0 : 1.0;
+            }
         }
     }
 }
 
-- (void) sendTouchEventWithTimestamp:(ulong)timeStamp
+- (void)sendTouchEventWithTimestamp:(ulong)timeStamp
 {
-    // Send touch event synchronously
     QIOSIntegration *iosIntegration = QIOSIntegration::instance();
-    QWindowSystemInterface::handleTouchEvent(m_qioswindow->window(), timeStamp, iosIntegration->touchDevice(), m_activeTouches.values());
-    QWindowSystemInterface::flushWindowSystemEvents();
+    if (!static_cast<QUIWindow *>(self.window).sendingEvent) {
+        // The event is likely delivered as part of delayed touch delivery, via
+        // _UIGestureEnvironmentSortAndSendDelayedTouches, due to one of the two
+        // _UISystemGestureGateGestureRecognizer instances on the top level window
+        // having its delaysTouchesBegan set to YES. During this delivery, it's not
+        // safe to spin up a recursive event loop, as our calling function is not
+        // reentrant, so any gestures used by the recursive code, e.g. a native
+        // alert dialog, will fail to recognize. To be on the safe side, we deliver
+        // the event asynchronously.
+        QWindowSystemInterface::handleTouchEvent<QWindowSystemInterface::AsynchronousDelivery>(
+            m_qioswindow->window(), timeStamp, iosIntegration->touchDevice(), m_activeTouches.values());
+    } else {
+        QWindowSystemInterface::handleTouchEvent<QWindowSystemInterface::SynchronousDelivery>(
+            m_qioswindow->window(), timeStamp, iosIntegration->touchDevice(), m_activeTouches.values());
+    }
 }
 
 - (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
@@ -361,7 +482,8 @@
     // We do this by assuming that there are no cases where a
     // sub-set of the active touch events are intentionally cancelled.
 
-    if (touches && (static_cast<NSInteger>([touches count]) != m_activeTouches.count()))
+    NSInteger count = static_cast<NSInteger>([touches count]);
+    if (count != 0 && count != m_activeTouches.count())
         qWarning("Subset of active touches cancelled by UIKit");
 
     m_activeTouches.clear();
@@ -369,22 +491,80 @@
 
     NSTimeInterval timestamp = event ? event.timestamp : [[NSProcessInfo processInfo] systemUptime];
 
-    // Send cancel touch event synchronously
     QIOSIntegration *iosIntegration = static_cast<QIOSIntegration *>(QGuiApplicationPrivate::platformIntegration());
     QWindowSystemInterface::handleTouchCancelEvent(m_qioswindow->window(), ulong(timestamp * 1000), iosIntegration->touchDevice());
-    QWindowSystemInterface::flushWindowSystemEvents();
+}
+
+- (int)mapPressTypeToKey:(UIPress*)press
+{
+    switch (press.type) {
+    case UIPressTypeUpArrow: return Qt::Key_Up;
+    case UIPressTypeDownArrow: return Qt::Key_Down;
+    case UIPressTypeLeftArrow: return Qt::Key_Left;
+    case UIPressTypeRightArrow: return Qt::Key_Right;
+    case UIPressTypeSelect: return Qt::Key_Select;
+    case UIPressTypeMenu: return Qt::Key_Menu;
+    case UIPressTypePlayPause: return Qt::Key_MediaTogglePlayPause;
+    }
+    return Qt::Key_unknown;
+}
+
+- (bool)processPresses:(NSSet *)presses withType:(QEvent::Type)type {
+    // Presses on Menu button will generate a Menu key event. By default, not handling
+    // this event will cause the application to return to Headboard (tvOS launcher).
+    // When handling the event (for example, as a back button), both press and
+    // release events must be handled accordingly.
+
+    bool handled = false;
+    for (UIPress* press in presses) {
+        int key = [self mapPressTypeToKey:press];
+        if (key == Qt::Key_unknown)
+            continue;
+        if (QWindowSystemInterface::handleKeyEvent(m_qioswindow->window(), type, key, Qt::NoModifier))
+            handled = true;
+    }
+
+    return handled;
+}
+
+- (void)pressesBegan:(NSSet<UIPress *> *)presses withEvent:(UIPressesEvent *)event
+{
+    if (![self processPresses:presses withType:QEvent::KeyPress])
+        [super pressesBegan:presses withEvent:event];
+}
+
+- (void)pressesChanged:(NSSet<UIPress *> *)presses withEvent:(UIPressesEvent *)event
+{
+    if (![self processPresses:presses withType:QEvent::KeyPress])
+        [super pressesChanged:presses withEvent:event];
+}
+
+- (void)pressesEnded:(NSSet<UIPress *> *)presses withEvent:(UIPressesEvent *)event
+{
+    if (![self processPresses:presses withType:QEvent::KeyRelease])
+        [super pressesEnded:presses withEvent:event];
 }
 
 - (BOOL)canPerformAction:(SEL)action withSender:(id)sender
 {
+#ifndef Q_OS_TVOS
     // Check first if QIOSMenu should handle the action before continuing up the responder chain
     return [QIOSMenu::menuActionTarget() targetForAction:action withSender:sender] != 0;
+#else
+    Q_UNUSED(action)
+    Q_UNUSED(sender)
+    return false;
+#endif
 }
 
 - (id)forwardingTargetForSelector:(SEL)selector
 {
     Q_UNUSED(selector)
+#ifndef Q_OS_TVOS
     return QIOSMenu::menuActionTarget();
+#else
+    return nil;
+#endif
 }
 
 @end
@@ -394,7 +574,7 @@
 - (QWindow *)qwindow
 {
     if ([self isKindOfClass:[QUIView class]]) {
-        if (QIOSWindow *w = static_cast<QUIView *>(self)->m_qioswindow)
+        if (QT_PREPEND_NAMESPACE(QIOSWindow) *w = static_cast<QUIView *>(self)->m_qioswindow)
             return w->window();
     }
     return nil;
@@ -419,7 +599,25 @@
     return nil;
 }
 
+- (UIEdgeInsets)qt_safeAreaInsets
+{
+#if QT_DARWIN_PLATFORM_SDK_EQUAL_OR_ABOVE(__MAC_NA, 110000, 110000, __WATCHOS_NA)
+    if (__builtin_available(iOS 11, tvOS 11, *))
+        return self.safeAreaInsets;
+#endif
+
+    // Fallback for iOS < 11
+    UIEdgeInsets safeAreaInsets = UIEdgeInsetsZero;
+    CGPoint topInset = [self convertPoint:CGPointMake(0, self.viewController.topLayoutGuide.length) fromView:nil];
+    CGPoint bottomInset = [self convertPoint:CGPointMake(0, self.viewController.bottomLayoutGuide.length) fromView:nil];
+    safeAreaInsets.top = topInset.y;
+    safeAreaInsets.bottom = bottomInset.y;
+    return safeAreaInsets;
+}
+
 @end
 
+#ifndef QT_NO_ACCESSIBILITY
 // Include category as an alternative to using -ObjC (Apple QA1490)
 #include "quiview_accessibility.mm"
+#endif
