@@ -1,31 +1,26 @@
 /****************************************************************************
 **
-** Copyright (C) 2014 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the test suite of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL21$
+** $QT_BEGIN_LICENSE:GPL-EXCEPT$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia. For licensing terms and
-** conditions see http://qt.digia.com/licensing. For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 or version 3 as published by the Free
-** Software Foundation and appearing in the file LICENSE.LGPLv21 and
-** LICENSE.LGPLv3 included in the packaging of this file. Please review the
-** following information to ensure the GNU Lesser General Public License
-** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights. These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 3 as published by the Free Software
+** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -42,6 +37,7 @@
 #include <qdebug.h>
 #include <qstring.h>
 #include <qshortcut.h>
+#include <qscreen.h>
 
 class AccelForm;
 QT_BEGIN_NAMESPACE
@@ -71,7 +67,8 @@ public:
         TriggerSlot4,
         TriggerSlot5,
         TriggerSlot6,
-        TriggerSlot7
+        TriggerSlot7,
+        SendKeyEvent
     };
 
     enum Result {
@@ -83,6 +80,7 @@ public:
         Slot5Triggered,
         Slot6Triggered,
         Slot7Triggered,
+        SentKeyEvent,
         Ambiguous
     } currentResult;
 
@@ -103,10 +101,12 @@ public slots:
     void ambigSlot7() { currentResult = Ambiguous; ambigResult = Slot7Triggered; }
     void statusMessage( const QString& message ) { sbText = message; }
     void shortcutDestroyed(QObject* obj);
+    void sendKeyEvent() { sendKeyEvents(edit, Qt::CTRL + Qt::Key_B, 0); currentResult = SentKeyEvent; }
 
 public slots:
     void initTestCase();
     void cleanupTestCase();
+    void cleanup() { QCOMPARE(QApplication::topLevelWidgets().size(), 1); }
 
 private slots:
     void number_data();
@@ -119,6 +119,7 @@ private slots:
     void keypressConsumption();
     void unicodeCompare();
     void context();
+    void duplicatedShortcutOverride();
 
 protected:
     static Qt::KeyboardModifiers toButtons( int key );
@@ -177,12 +178,13 @@ protected:
             QKeyEvent *ke = static_cast<QKeyEvent*>(e);
             if (ke->modifiers() && ke->key() > Qt::Key_Any
                 && ke->key() < Qt::Key_ydiaeresis) {
+                const QChar c = QLatin1Char(char(ke->key()));
                 if (ke->modifiers() == Qt::ControlModifier)
-                    insertPlainText(QString("<Ctrl+%1>").arg(char(ke->key())));
+                    insertPlainText(QLatin1String("<Ctrl+") + c + QLatin1Char('>'));
                 else if (ke->modifiers() == Qt::AltModifier)
-                    insertPlainText(QString("<Alt+%1>").arg(char(ke->key())));
+                    insertPlainText(QLatin1String("<Alt+") + c + QLatin1Char('>'));
                 else if (ke->modifiers() == Qt::ShiftModifier)
-                    insertPlainText(QString("<Shift+%1>").arg(char(ke->key())));
+                    insertPlainText(QLatin1String("<Shift+") + c + QLatin1Char('>'));
                 return true;
             }
         }
@@ -699,14 +701,14 @@ void tst_QShortcut::disabledItems()
     sendKeyEvents( Qt::CTRL+Qt::Key_M, 0 );
     QCOMPARE( currentResult, NoResult );
     if (over_330)
-        QCOMPARE( sbText, QString::null );
+        QCOMPARE( sbText, QString() );
 
     currentResult = NoResult;
     sendKeyEvents( Qt::CTRL+Qt::Key_K, 0 );
     sendKeyEvents( Qt::CTRL+Qt::Key_L, 0 );
     QCOMPARE( currentResult, Slot1Triggered );
     if (over_330)
-        QCOMPARE( sbText, QString::null );
+        QCOMPARE( sbText, QString() );
 #endif
     clearAllShortcuts();
     cut1 = 0;
@@ -978,6 +980,19 @@ void tst_QShortcut::keypressConsumption()
     QVERIFY(edit->toPlainText().endsWith("<Ctrl+I>a"));
 
     clearAllShortcuts();
+    edit->clear();
+    QCOMPARE(edit->toPlainText().size(), 0);
+
+    setupShortcut(edit, "first", SendKeyEvent, "Ctrl+A");
+
+    // Verify reentrancy when a non-shortcut is triggered as part
+    // of shortcut processing.
+    currentResult = NoResult;
+    ambigResult = NoResult;
+    sendKeyEvents(edit, Qt::CTRL + Qt::Key_A, 0);
+    QCOMPARE(currentResult, SentKeyEvent);
+    QCOMPARE(ambigResult, NoResult);
+    QCOMPARE(edit->toPlainText(), QString(QString("<Ctrl+B>")));
 }
 
 // ------------------------------------------------------------------
@@ -1082,6 +1097,36 @@ void tst_QShortcut::context()
     clearAllShortcuts();
 }
 
+// QTBUG-38986, do not generate duplicated QEvent::ShortcutOverride in event processing.
+class OverrideCountingWidget : public QWidget
+{
+public:
+    OverrideCountingWidget(QWidget *parent = 0) : QWidget(parent), overrideCount(0) {}
+
+    int overrideCount;
+
+    bool event(QEvent *e) Q_DECL_OVERRIDE
+    {
+        if (e->type() == QEvent::ShortcutOverride)
+            overrideCount++;
+        return QWidget::event(e);
+    }
+};
+
+void tst_QShortcut::duplicatedShortcutOverride()
+{
+    OverrideCountingWidget w;
+    w.setWindowTitle(Q_FUNC_INFO);
+    w.resize(200, 200);
+    w.move(QGuiApplication::primaryScreen()->availableGeometry().center() - QPoint(100, 100));
+    w.show();
+    QApplication::setActiveWindow(&w);
+    QVERIFY(QTest::qWaitForWindowActive(&w));
+    QTest::keyPress(w.windowHandle(), Qt::Key_A);
+    QCoreApplication::processEvents();
+    QCOMPARE(w.overrideCount, 1);
+}
+
 // ------------------------------------------------------------------
 // Element Testing helper functions ---------------------------------
 // ------------------------------------------------------------------
@@ -1149,9 +1194,12 @@ QShortcut *tst_QShortcut::setupShortcut(QWidget *parent, const char *name, int t
         normal = SLOT(slotTrig7());
         ambig  = SLOT(ambigSlot7());
         break;
+    case SendKeyEvent:
+        normal = SLOT(sendKeyEvent());
     }
     connect(cut, SIGNAL(activated()), this, normal);
-    connect(cut, SIGNAL(activatedAmbiguously()), this, ambig);
+    if (ambig)
+        connect(cut, SIGNAL(activatedAmbiguously()), this, ambig);
     connect(cut, SIGNAL(destroyed(QObject*)), this, SLOT(shortcutDestroyed(QObject*)));
     shortcuts.append(cut);
     return cut;
@@ -1159,7 +1207,8 @@ QShortcut *tst_QShortcut::setupShortcut(QWidget *parent, const char *name, int t
 
 void tst_QShortcut::shortcutDestroyed(QObject* obj)
 {
-    shortcuts.removeAll(static_cast<QShortcut *>(obj));
+    shortcuts.erase(std::remove(shortcuts.begin(), shortcuts.end(), obj),
+                    shortcuts.end());
 }
 
 void tst_QShortcut::sendKeyEvents(int k1, QChar c1, int k2, QChar c2, int k3, QChar c3, int k4, QChar c4)
@@ -1226,7 +1275,7 @@ void tst_QShortcut::testElement()
         setupShortcut(testWidget, txt, k1, k2, k3, k4);
     } else {
         sendKeyEvents(k1, c1, k2, c2, k3, c3, k4, c4);
-        QCOMPARE(currentResult, result);
+        QCOMPARE((int)currentResult, (int)result);
     }
 }
 
